# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging
import re

from climetlab.decorators import normalize, normalize_grib_keys
from climetlab.utils.humanize import list_to_human

LOG = logging.getLogger(__name__)

# Make sure the

ACCUMULATIONS = {("tp", 2): {"productDefinitionTemplateNumber": 8}}


_ORDER = (
    "edition",
    "setLocalDefinition",
    "typeOfGeneratingProcess",
    "productDefinitionTemplateNumber",
)

NOT_IN_EDITION_1 = (
    "productDefinitionTemplateNumber",
    "typeOfGeneratingProcess",
)


ORDER = {}
for i, k in enumerate(_ORDER):
    ORDER[k] = i


def order(key):
    ORDER.setdefault(key, len(ORDER))
    return ORDER[key]


class Combined:
    def __init__(self, handle, metadata):
        self.handle = handle
        self.metadata = metadata

    def __contains__(self, key):
        # return key in self.metadata or key in self.handle
        raise NotImplementedError()

    def __getitem__(self, key):
        if key in self.metadata:
            return self.metadata[key]
        return self.handle.get(key)


class GribOutput:
    def __init__(self, filename, split_output=False, template=None, **kwargs):
        self._files = {}
        self.filename = filename

        if split_output:
            self.split_output = re.findall(r"\{(.*?)\}", self.filename)
        else:
            self.split_output = None

        self.template = template
        self._bbox = {}
        self.kwargs = kwargs

    @normalize_grib_keys
    @normalize("date", "date")
    def _normalize_kwargs_names(self, **kwargs):
        return kwargs

    def f(self, handle):
        if self.split_output:
            path = self.filename.format(**{k: handle.get(k) for k in self.split_output})
        else:
            path = self.filename

        if path not in self._files:
            self._files[path] = open(path, "wb")
        return self._files[path], path

    def write(
        self,
        values,
        check_nans=False,
        metadata={},
        template=None,
        **kwargs,
    ):
        # Make a copy as we may modify it
        md = self._normalize_kwargs_names(**self.kwargs)
        md.update(self._normalize_kwargs_names(**metadata))
        md.update(self._normalize_kwargs_names(**kwargs))

        metadata = md

        compulsary = (("date", "referenceDate"), ("param", "paramId", "shortName"))

        if template is None:
            template = self.template

        if template is None:
            handle = self.handle_from_metadata(values, metadata, compulsary)
        else:
            handle = template.handle.clone()

        # print("->", metadata)
        self.update_metadata(handle, metadata, compulsary)
        # print("<-", metadata)

        if check_nans and values is not None:
            import numpy as np

            if np.isnan(values).any():
                # missing_value = np.finfo(values.dtype).max
                missing_value = 9999
                values = np.nan_to_num(values, nan=missing_value)
                metadata["missingValue"] = missing_value
                metadata["bitmapPresent"] = 1

        metadata = {
            k: v for k, v in sorted(metadata.items(), key=lambda x: order(x[0]))
        }

        if str(metadata.get("edition")) == "1":
            for k in NOT_IN_EDITION_1:
                metadata.pop(k, None)

        LOG.debug("GribOutput.metadata %s", metadata)

        for k, v in metadata.items():
            handle.set(k, v)

        if values is not None:
            handle.set_values(values)

        # Set values will set generatingProcessIdentifier to 255
        if "generatingProcessIdentifier" in metadata:
            handle.set(
                "generatingProcessIdentifier", metadata["generatingProcessIdentifier"]
            )

        file, path = self.f(handle)
        handle.write(file)

        return handle, path

    def close(self):
        for f in self._files.values():
            f.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trace):
        self.close()

    def update_metadata(self, handle, metadata, compulsary):
        # TODO: revisit that logic
        combined = Combined(handle, metadata)

        if "step" in metadata or "endStep" in metadata:
            if combined["type"] == "an":
                metadata["type"] = "fc"

        if "time" in metadata:  # TODO, use a normalizer
            try:
                time = int(metadata["time"])
                if time < 100:
                    metadata["time"] = time * 100
            except ValueError:
                pass

        if "time" not in metadata and "date" in metadata:
            date = metadata["date"]
            metadata["time"] = date.hour * 100 + date.minute

        if "date" in metadata:
            if isinstance(metadata["date"], datetime.datetime):
                date = metadata["date"]
                metadata["date"] = date.year * 10000 + date.month * 100 + date.day
            else:
                metadata["date"] = int(metadata["date"])

        if "stream" not in metadata:
            if "number" in metadata:
                metadata["stream"] = "enfo"
                metadata.setdefault("type", "pf")

        if "number" in metadata:
            compulsary += ("numberOfForecastsInEnsemble",)
            productDefinitionTemplateNumber = {"tp": 11}
            metadata["productDefinitionTemplateNumber"] = (
                productDefinitionTemplateNumber.get(handle.get("shortName"), 1)
            )

        if metadata.get("type") in ("pf", "cf"):
            metadata.setdefault("typeOfGeneratingProcess", 4)

        if "levelist" in metadata:
            metadata.setdefault("levtype", "pl")

        if "param" in metadata:
            param = metadata.pop("param")
            try:
                metadata["paramId"] = int(param)
            except ValueError:
                metadata["shortName"] = param

        # levtype is a readOnly key in ecCodes >= 2.33.0
        levtype_remap = {
            "pl": "isobaricInhPa",
            "ml": "hybrid",
            "pt": "theta",
            "pv": "potentialVorticity",
            "sfc": "surface",
        }
        if "levtype" in metadata:
            v = metadata.pop("levtype")
            metadata["typeOfLevel"] = levtype_remap[v]

    def handle_from_metadata(self, values, metadata, compulsary):
        from .codes import CodesHandle  # Lazy loading of eccodes

        if len(values.shape) == 1:
            sample = self._gg_field(values, metadata)
        elif len(values.shape) == 2:
            sample = self._ll_field(values, metadata)
        else:
            raise ValueError(
                f"Invalid shape {values.shape} for GRIB, must be 1 or 2 dimension "
            )

        metadata.setdefault("bitsPerValue", 16)
        metadata["scanningMode"] = 0

        metadata.update(
            ACCUMULATIONS.get(
                (
                    metadata.get("paramId"),
                    metadata.get("edition", 2),
                ),
                {},
            )
        )
        metadata.update(
            ACCUMULATIONS.get(
                (
                    metadata.get("shortName"),
                    metadata.get("edition", 2),
                ),
                {},
            )
        )

        if (
            "class" in metadata
            or "type" in metadata
            or "stream" in metadata
            or "expver" in metadata
        ):
            # MARS labelling
            metadata["setLocalDefinition"] = 1
            # metadata['grib2LocalSectionNumber'] = 1

        for check in compulsary:
            if not isinstance(check, tuple):
                check = [check]

            if not any(c in metadata for c in check):
                choices = list_to_human([f"'{c}'" for c in check], "or")
                raise ValueError(f"Please provide a value for {choices}.")

        LOG.debug("CodesHandle.from_sample(%s)", sample)
        return CodesHandle.from_sample(sample)

    def _ll_field(self, values, metadata):
        Nj, Ni = values.shape
        metadata["Nj"] = Nj
        metadata["Ni"] = Ni

        # We assume the scanning mode north->south, west->east
        west_east = 360 / Ni

        if Nj % 2 == 0:
            north_south = 180 / Nj
            adjust = north_south / 2
        else:
            north_south = 181 / Nj
            adjust = 0

        north = 90 - adjust
        south = -90 + adjust
        west = 0
        east = 360 - west_east

        metadata["iDirectionIncrementInDegrees"] = west_east
        metadata["jDirectionIncrementInDegrees"] = north_south

        metadata["latitudeOfFirstGridPointInDegrees"] = north
        metadata["latitudeOfLastGridPointInDegrees"] = south
        metadata["longitudeOfFirstGridPointInDegrees"] = west
        metadata["longitudeOfLastGridPointInDegrees"] = east

        edition = metadata.get("edition", 2)
        levtype = metadata.get("levtype")
        if levtype is None:
            if "levelist" in metadata:
                levtype = "pl"
            else:
                levtype = "sfc"

        return f"regular_ll_{levtype}_grib{edition}"

    def _gg_field(self, values, metadata):
        GAUSSIAN = {
            6114: (32, False),
            13280: (48, False),
            24572: (64, False),
            35718: (80, False),
            40320: (96, True),
            50662: (96, False),
            88838: (128, False),
            108160: (160, True),
            138346: (160, False),
            213988: (200, False),
            348528: (256, False),
            542080: (320, False),
            843490: (400, False),
            1373624: (512, False),
            2140702: (640, False),
            5447118: (1024, False),
            6599680: (1280, True),
            8505906: (1280, False),
            20696844: (2000, False),
        }

        n = len(values)
        if n not in GAUSSIAN:
            raise ValueError(f"Unsupported GAUSSIAN grid. Number of grid points {n:,}")
        N, octahedral = GAUSSIAN[n]

        if N not in self._bbox:
            import eccodes

            self._bbox[N] = max(eccodes.codes_get_gaussian_latitudes(N))

        metadata["latitudeOfFirstGridPointInDegrees"] = self._bbox[N]
        metadata["latitudeOfLastGridPointInDegrees"] = -self._bbox[N]
        metadata["longitudeOfFirstGridPointInDegrees"] = 0

        metadata["N"] = N
        if octahedral:
            half = list(range(20, 20 + N * 4, 4))
            pl = half + list(reversed(half))
            assert len(pl) == 2 * N, (len(pl), 2 * N)
            metadata["pl"] = pl
            metadata["longitudeOfLastGridPointInDegrees"] = 360 - max(pl) / 360
        else:
            # Assumed to be set properly in the sample
            # metadata["longitudeOfLastGridPointInDegrees"] = east
            pass

        edition = metadata.get("edition", 2)
        levtype = metadata.get("levtype")
        if levtype is None:
            if "levelist" in metadata:
                levtype = "pl"
            else:
                levtype = "sfc"

        if octahedral:
            return f"reduced_gg_{levtype}_grib{edition}"
        else:
            return f"reduced_gg_{levtype}_{N}_grib{edition}"


def new_grib_output(*args, **kwargs):
    return GribOutput(*args, **kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import atexit
import functools
import os
import threading
import time
from collections import defaultdict
from contextlib import contextmanager

from climetlab.utils.humanize import number, seconds

PROFILING = int(os.environ.get("CLIMETLAB_PROFILING", 0))
START = time.time()


@contextmanager
def timer(msg):
    start = time.time()
    print(f"{msg}: Starting.")
    try:
        yield None
    finally:
        print(f"{msg}: {time.time()-start}")


class C:
    n = 0
    now = 0
    elapsed = 0

    def start(self):
        self.n += 1
        if time is not None:
            self.now = time.time()

    def end(self):
        if time is not None:
            self.elapsed += time.time() - self.now

    def __repr__(self):
        m = self.n if self.n else 1
        return "calls: %s, elapsed: %s, average: %s" % (
            number(self.n),
            seconds(self.elapsed),
            seconds(self.elapsed / m),
        )


class Counter:
    def __init__(self, name):
        self.name = name
        self._c = C()
        self.threads = defaultdict(C)

    def __repr__(self):
        extra = ""
        if len(self.threads) > 1:
            extra = "\n threads:\n   %s" % (
                "\n   ".join(repr(t) for t in self.threads.values()),
            )
        return "COUNTER [%s], %s%s" % (
            self.name,
            self._c,
            extra,
        )

    def __enter__(self):
        if threading is not None:
            self._c.start()
            self.threads[threading.current_thread()].start()
            return self

    def __exit__(self, *args, **kwargs):
        if threading is not None:
            self._c.end()
            self.threads[threading.current_thread()].end()


COUNTERS = []


def print_counters():
    if PROFILING:
        for n in COUNTERS:
            print(n)
        print(f"ELAPSED [{seconds(time.time() - START)}]")


atexit.register(print_counters)


def call_counter(name=None):
    func = None
    if callable(name):
        func = name
        # TODO: Find the proper way of getting the
        # function full name.
        name = repr(func).split()[1]

    counter = Counter(name)
    COUNTERS.append(counter)

    def decorate(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with counter:
                return func(*args, **kwargs)

        return wrapper

    if func is not None:
        return decorate(func)

    return decorate



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os

from climetlab.utils import tqdm

from . import Reader
from . import reader as find_reader

LOG = logging.getLogger(__name__)


class ArchiveReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)

    def check(self, member):
        # A bit of paranoia

        if not member.isdir() and not member.isfile():
            LOG.warning(
                "Ignoring archive member '%s' because it is not a directory or a file",
                member.name,
            )
            return False

        if member.name.startswith("/"):
            LOG.warning(
                "Ignoring archive member '%s' because it is a full path",
                member.name,
            )
            return False

        if ".." in member.name:
            LOG.warning(
                "Ignoring archive member '%s' because contains '..'",
                member.name,
            )
            return False

        return True

    def mutate(self):
        if os.path.isdir(self.path):
            return find_reader(self.source, self.path).mutate()

        return self

    def expand(self, archive, members, **kwargs):
        def unpack(target, args):
            try:
                os.mkdir(target)
            except FileExistsError:
                pass

            for member in tqdm(iterable=members, total=len(members), leave=False):
                if not self.check(member):
                    continue
                archive.extract(member=member, path=target, **kwargs)

        self.path = self.cache_file(
            unpack,
            self.path,
            extension=".d",
            replace=self.path,
        )



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import numpy as np


class DataSet:
    """
    Class that wraps a xarray dataset to provide caching
    """

    def __init__(self, ds):
        assert ds is not None
        self._ds = ds
        self._cache = {}

    def __getitem__(self, name):
        return self._ds[name]

    def __getattr__(self, name):
        return getattr(self._ds, name)

    def bbox(self, variable):
        data_array = self[variable]
        dims = data_array.dims

        lat = dims[-2]
        lon = dims[-1]

        key = ("bbox", lat, lon)
        if key in self._cache:
            return self._cache[key]

        lats, lons = self.grid_points(variable)
        north = np.amax(lats)
        west = np.amin(lons)
        south = np.amin(lats)
        east = np.amax(lons)

        self._cache[key] = (north, west, south, east)
        return self._cache[key]

    def grid_points(self, variable):
        data_array = self[variable]
        dims = data_array.dims

        lat = dims[-2]
        lon = dims[-1]

        key = ("grid_points", lat, lon)
        if key in self._cache:
            return self._cache[key]

        if "latitude" in self._ds and "longitude" in self._ds:
            latitude = self._ds["latitude"]
            longitude = self._ds["longitude"]

            if latitude.dims == (lat, lon) and longitude.dims == (lat, lon):
                latitude = latitude.data
                longitude = longitude.data
                return latitude.flatten(), longitude.flatten()

        latitude = data_array[lat]
        longitude = data_array[lon]

        lat, lon = np.meshgrid(latitude.data, longitude.data)

        self._cache[key] = lat.flatten(), lon.flatten()
        return self._cache[key]

    def grid_points_xy(self, variable):
        data_array = self[variable]
        dims = data_array.dims

        lat = dims[-2]
        lon = dims[-1]

        latitude = data_array[lat].data
        longitude = data_array[lon].data

        print(latitude, longitude)

        lat, lon = np.meshgrid(latitude, longitude)

        return lat.flatten(), lon.flatten()
        # return self._cache[key]



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from climetlab.decorators import cached_method
from climetlab.indexing.database.json import JsonFileDatabase
from climetlab.readers.grib.index.db import FieldsetInFilesWithDBIndex


class FieldsetInFilesWithJsonIndex(FieldsetInFilesWithDBIndex):
    DBCLASS = JsonFileDatabase

    @cached_method
    def _lookup_parts(self):
        return self.db.lookup_parts()

    def part(self, n):
        return self._lookup_parts()[n]

    def number_of_parts(self):
        return len(self._lookup_parts())



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab import load_source

from .multi import MultiSource


class MultiUrl(MultiSource):
    def __init__(self, urls, *args, filter=None, merger=None, force=None, **kwargs):
        if not isinstance(urls, (list, tuple)):
            urls = [urls]

        # if filter is not None:
        #     urls = [url for url in urls if filter(url)]

        assert len(urls)

        sources = [
            load_source(
                "url",
                url,
                filter=filter,
                merger=merger,
                force=force,
                # Load lazily so we can do parallel downloads
                lazily=True,
            )
            for url in sorted(urls)
        ]

        super().__init__(sources, filter=filter, merger=merger)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from importlib import import_module

LOG = logging.getLogger(__name__)

SERIALISATION = {}


def register_serialisation(cls, serialise, create):
    fullname = (cls.__module__, cls.__qualname__)
    SERIALISATION[fullname] = (serialise, create)


def serialise_state(obj):
    fullname = (obj.__class__.__module__, obj.__class__.__qualname__)
    LOG.info("serialise %s", fullname)
    return (fullname, SERIALISATION[fullname][0](obj))


def deserialise_state(state):
    fullname, args = state
    LOG.info("deserialise %s %s", fullname, args)
    # warnings.warn('deserialise %s %s', fullname, args)
    # TODO: if KeyError, maybe the function is not imported
    module, _ = fullname
    import_module(module)

    create = SERIALISATION[fullname][1]
    return create(args)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os


def get_version():
    version_file = os.path.join(os.path.dirname(__file__), "version")
    with open(version_file, "r") as f:
        version = f.readlines()
        version = version[0]
        version = version.strip()
    return version


__version__ = get_version()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import tempfile


class TmpFile:
    """The TmpFile objets are designed to be used for temporary files.
    It ensures that the file is unlinked when the object is
    out-of-scope (with __del__).

    Parameters
    ----------
    path : str
        Actual path of the file.
    """

    def __init__(self, path: str):
        self.path = path

    def __del__(self):
        self.cleanup()

    def __enter__(self):
        return self.path

    def __exit__(self, *args, **kwargs):
        self.cleanup()

    def cleanup(self):
        if self.path is not None and os is not None:
            os.unlink(self.path)
        self.path = None


def temp_file(extension=".tmp") -> TmpFile:
    """Create a temporary file with the given extension.

    Parameters
    ----------
    extension : str, optional
        By default ".tmp"

    Returns
    -------
    TmpFile
    """

    fd, path = tempfile.mkstemp(suffix=extension)
    os.close(fd)
    return TmpFile(path)


class TmpDirectory(tempfile.TemporaryDirectory):
    @property
    def path(self):
        return self.name


def temp_directory():
    return TmpDirectory()


class TmpEnv:
    """
    A context manager that temporarily sets environment variables.

    Usage:
    >>> import os
    >>> with TmpEnv(CLIMETLAB_TESTS_FOO='123', CLIMETLAB_TESTS_BAR='456'):
    ...     print(os.environ['CLIMETLAB_TESTS_FOO'])
    ...     print(os.environ['CLIMETLAB_TESTS_BAR'])
    123
    456
    >>> print(os.environ.get('CLIMETLAB_TESTS_FOO'))
    None
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.previous = {}

    def __enter__(self):
        for key, value in self.kwargs.items():
            self.previous[key] = os.environ.get(key)
            os.environ[key] = str(value)

    def __exit__(self, type, value, traceback):
        for key in self.kwargs.keys():
            if self.previous.get(key) is not None:
                os.environ[key] = self.previous[key]
            else:
                del os.environ[key]


def temp_env(**kwargs):
    return TmpEnv(**kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import stat
from zipfile import ZipFile

from climetlab import load_source

from .archive import ArchiveReader
from .csv import CSVReader


class InfoWrapper:
    """
    A class so that ZipInfo has the same interface as TarInfo
    """

    def __init__(self, member):
        self.member = member
        self.file_or_directory = True
        # See https://stackoverflow.com/questions/35782941/archiving-symlinks-with-python-zipfile
        if member.create_system == 3:  # Unix
            unix_mode = member.external_attr >> 16
            if unix_mode & 0o100000 == 0:
                # if mode is 0, starts with '?' in zipinfo, set it to 1
                unix_mode = unix_mode | 0o100000
            self.file_or_directory = stat.S_ISDIR(unix_mode) or stat.S_ISREG(unix_mode)

    @property
    def name(self):
        return self.member.filename

    def isdir(self):
        return self.file_or_directory and self.member.filename.endswith("/")

    def isfile(self):
        return self.file_or_directory and not self.isdir()


class ZIPReader(ArchiveReader):
    def __init__(self, source, path):
        super().__init__(source, path)

        self._mutate = None

        with ZipFile(path, "r") as zip:
            members = zip.infolist()

            if len(members) == 1:
                _, ext = os.path.splitext(members[0].filename)
                if ext in (".csv",):
                    self._mutate = CSVReader(source, path, compression="zip")
                    return  # Pandas can read zipped files directly

            if ".zattrs" in members:
                return  # Zarr can read zipped files directly

            self.expand(zip, members)

    def check(self, member):
        return super().check(InfoWrapper(member))

    def mutate(self):
        if self._mutate:
            return self._mutate

        return super().mutate()

    def mutate_source(self):
        # zarr can read data from a zip file
        if ".zattrs" in self._content:
            return load_source("zarr", self.path)

        return None


EXTENSIONS_TO_SKIP = (".npz",)  # Numpy arrays


def reader(source, path, magic=None, deeper_check=False):
    if magic is None:  # Bypass check and force
        return ZIPReader(source, path)

    _, extension = os.path.splitext(path)

    if magic[:4] == b"PK\x03\x04" and extension not in EXTENSIONS_TO_SKIP:
        return ZIPReader(source, path)



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import torch

from .utils import as_numpy_func, default_merger


class TorchDataset(torch.utils.data.Dataset):
    def __init__(self, owner, elements):
        self.owner = owner
        self.elements = elements

    def __len__(self):
        return len(self.owner)

    def __getitem__(self, i):
        return self.elements.get_torch_item(i)
        # return self.owner.get_torch_item(i)


def build_data_specs(*args, output, **kwargs):
    return {
        "x,y": XYDataIO,
        "x": XDataIO,
        # TODO: einops style:
        #  "(bs, lat * lon, param * lev)":
        #  "(bs, lat * lon, param_lev)":
        #  "bs, (lat lon), (param lev)":
        #  "bs, (param lev), (lat lon)":
        # plev vs lev ?
        # param_plev ?
    }[output](*args, **kwargs)


class DataIO:
    def __init__(self, *args, owner=None, **kwargs):
        assert not args, "only keywords args are supported"
        self.owner = owner

        self.args = args
        self.kwargs = kwargs

    def merge_elements(self, elements, merger):
        funcs = [elt.func for name, elt in elements.items()]
        return merger(*funcs)

    def get_torch_item(self, key):
        raise NotImplementedError()


class XDataIO(DataIO):
    def __init__(self, *args, owner=None, **kwargs):
        super().__init__(*args, owner=owner, **kwargs)

        self._features = {}

        self.feature_options = kwargs.get("features_options", {})
        self.features_merger = kwargs.get("features_merger", default_merger)

        for name in kwargs["features"]:
            options = self.feature_options.get(name, {})
            self._features[name] = Element(name, owner, options).mutate()

        self.func = self.merge_elements(self._features, self.features_merger)

    def get_torch_item(self, key):
        return self.func(key)


class XYDataIO(DataIO):
    def __init__(self, *args, owner=None, **kwargs):
        super().__init__(*args, owner=owner, **kwargs)

        self._features = {}
        self._targets = {}

        self.feature_options = kwargs.get("features_options", {})
        self.target_options = kwargs.get("target_options", {})
        self.features_merger = kwargs.get("features_merger", default_merger)
        self.targets_merger = kwargs.get("targets_merger", default_merger)

        for name in kwargs["features"]:
            options = self.feature_options.get(name, {})
            self._features[name] = Element(name, owner, options).mutate()

        for name in kwargs["targets"]:
            options = self.feature_options.get(name, {})
            self._targets[name] = Element(name, owner, options).mutate()

        self.func = self.merge_elements(self._features, self.features_merger)
        self.target_func = self.merge_elements(self._targets, self.targets_merger)

    def get_torch_item(self, key):
        return self.func(key), self.target_func(key)


class Element:
    def __init__(self, string, dataset_or_source, options):
        self.dataset_or_source = dataset_or_source
        self._source = None
        self._init_string = string
        self.options = options

        self.name = string
        assert self.name is not None

    @property
    def source(self):
        if self._source is None:
            self._source = self.dataset_or_source.build_source_for_element(self)
        return self._source

    def get_item(self, i):
        return self.source[i]

    def mutate(self):
        return self

    @property
    def func(self):
        return as_numpy_func(self.source, self.options)  # refactor HERE TODO



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import itertools

import climetlab as cml
from climetlab import Source
from climetlab.decorators import normalize


class OperAccumulations(Source):
    def __init__(self, *args, **kwargs):
        request = {}
        for a in args:
            request.update(self.requests(**a))
        request = self.requests(**kwargs)

        param = request["param"]

        if not isinstance(param, (list, tuple)):
            param = [param]

        for p in param:
            assert p in ["cp", "lsp", "tp"], p

        user_dates = request["date"]
        user_times = request["time"]

        requests = {
            "oper": {"dates": set(), "times": set()},
            "scda": {"dates": set(), "times": set()},
        }

        user_step = 6
        requested = set()

        for user_date, user_time in itertools.product(user_dates, user_times):
            assert isinstance(user_date, datetime.datetime), (
                type(user_date),
                user_dates,
                user_times,
            )
            assert isinstance(user_time, int), (type(user_time), user_dates, user_times)
            assert user_time in [0, 6, 12, 18], user_time

            when = user_date + datetime.timedelta(hours=user_time)

            requested.add(when)

            when -= datetime.timedelta(hours=user_step)
            date = datetime.datetime(when.year, when.month, when.day)
            time = when.hour

            stream = {0: "oper", 6: "scda", 12: "oper", 18: "scda"}[time]
            requests[stream]["dates"].add(date)
            requests[stream]["times"].add(time)
            print(requests)

        dataset = dict(oper=cml.load_source("empty"), scda=cml.load_source("empty"))

        for stream in ["oper", "scda"]:
            dates = sorted(requests[stream]["dates"])
            times = sorted(requests[stream]["times"])

            if not dates and not times:
                continue

            assert dates, (stream, dates, times)

            oper_request = dict(**request)

            oper_request.update(
                {
                    "class": "od",
                    "type": "fc",
                    "levtype": "sfc",
                    "stream": stream,
                    "date": [d.strftime("%Y-%m-%d") for d in dates],
                    "time": sorted(times),
                    "step": user_step,
                }
            )

            ds = cml.load_source("mars", **oper_request)
            index = [d.valid_datetime() in requested for d in ds]
            dataset[stream] = ds[index]

        self.ds = dataset["oper"] + dataset["scda"]

    def mutate(self):
        return self.ds

    @normalize("date", "date-list(datetime.datetime)")
    @normalize("time", "int-list")
    @normalize("area", "bounding-box(list)")
    def requests(self, **kwargs):
        result = dict(**kwargs)

        return result


source = OperAccumulations



#!/usr/bin/env python
#
# (C) Copyright 2012- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#


import datetime
import itertools
from collections import defaultdict
from copy import copy
from functools import cmp_to_key

from dateutil.parser import parse as parse_dates


class Interval(object):
    def __init__(self, start, end):
        self.start = start
        self.end = end
        if not isinstance(start, datetime.date) and not isinstance(end, datetime.date):
            self.one = datetime.timedelta(hours=1)
            raise NotImplementedError("Interval with time not yet supported")
        else:
            self.one = datetime.timedelta(days=1)
        assert start <= end
        assert isinstance(start, datetime.date)
        assert isinstance(end, datetime.date)

    def count(self):
        return (self.end - self.start).days + 1

    def split(self, dates):
        result = []

        bounds = set(dates)
        bounds.discard(self.start)
        bounds.discard(self.end)

        s = self.start
        e = self.end
        for b in sorted(bounds):
            if b >= s and b <= e:
                result.append(self.__class__(s, b))
                s = b + self.one

        result.append(self.__class__(s, e))
        return result

    def overlaps(self, other):
        """Returns the union of two intervals if they overlap, else None"""
        s1, e1 = self.start, self.end + self.one
        s2, e2 = other.start, other.end + self.one
        s = max(s1, s2)
        e = min(e1, e2)
        if s <= e:
            return self.__class__(min(s1, s2), max(e1, e2) - self.one)
        else:
            return None

    def intersects(self, other):
        """Returns the intersection of two intervals if they overlap, else None"""
        s1, e1 = self.start, self.end + self.one
        s2, e2 = other.start, other.end + self.one
        s = max(s1, s2)
        e = min(e1, e2)
        if s <= e:
            return self.__class__(max(s1, s2), min(e1, e2) - self.one)
        else:
            return None

    @classmethod
    def expand(cls, lst):
        for i in lst:
            d = i.start
            while d <= i.end:
                yield d.date()
                d = d + datetime.timedelta(days=1)

    @classmethod
    def intersection(cls, list1, list2):
        result = []
        for i1 in list1:
            for i2 in list2:
                result.append(i1.intersects(i2))

        return Interval.join([r for r in result if r is not None])

    @classmethod
    def join(cls, intervals):
        result = sorted(intervals)

        more = True
        while more:
            more = False

            for i in range(len(result)):
                if result[i] is not None:
                    for j in range(i + 1, len(result)):
                        o = result[i].overlaps(result[j])
                        if o is not None:
                            result[i] = o
                            result[j] = None
                            more = True

            result = [r for r in result if r is not None]

        return tuple(result)

    def __repr__(self):
        if isinstance(self.start, datetime.date):
            if self.start == self.end:
                return self.start.strftime("%Y-%m-%d")

            return "%s/%s" % (
                self.start.strftime("%Y-%m-%d"),
                self.end.strftime("%Y-%m-%d"),
            )

        if self.start == self.end:
            return self.start.isoformat()
        return "%s/%s" % (self.start.isoformat(), self.end.isoformat())

    def __lt__(self, other):
        return (self.start, self.end) < (other.start, other.end)

    def __le__(self, other):
        return (self.start, self.end) <= (other.start, other.end)

    def __gt__(self, other):
        return (self.start, self.end) > (other.start, other.end)

    def __ge__(self, other):
        return (self.start, self.end) >= (other.start, other.end)

    def __eq__(self, other):
        return (self.start, self.end) == (other.start, other.end)

    def __ne__(self, other):
        return (self.start, self.end) != (other.start, other.end)

    def __hash__(self):
        return hash((self.start, self.end))


def _cleanup(x):
    if isinstance(x, (list, tuple)):
        return [_cleanup(a) for a in x]

    if isinstance(x, dict):
        return {_cleanup(k): _cleanup(v) for k, v in x.items()}

    if isinstance(x, (str, int, float)):
        return x

    return str(repr(x))


def _to_hashable(x):
    assert isinstance(x, dict)
    return tuple((k, v) for k, v in sorted(x.items()))


def _from_hashable(x):
    assert isinstance(x, tuple)
    return {k: v for k, v in x}


def _as_tuple(t):
    if isinstance(t, tuple):
        return t
    if isinstance(t, list) and len(t) == 1:
        return _as_tuple(t[0])
    if isinstance(t, (list, set)):
        return tuple(t)
    return (t,)


def _as_interval(interval):
    if not isinstance(interval, (list, tuple)):
        interval = [interval]
    result = []
    for t in interval:
        if isinstance(t, Interval):
            result.append(t)
            continue

        if isinstance(t, datetime.date):
            result.append(Interval(t, t))
            continue

        bits = t.split("/")
        assert len(bits) in (1, 2)
        if len(bits) == 1:
            start = end = bits[0]
        else:
            start = bits[0]
            end = bits[1]
        start = parse_dates(start)
        end = parse_dates(end)
        result.append(Interval(start, end))
    return result


class Tree:
    def __init__(self, values=None, intervals=None):
        self._values = {} if values is None else values
        self._children = []
        self._unique_values = None
        self._flatten = None
        self._intervals = set() if intervals is None else intervals

    def _add_child(self, child):
        self._children.append(child.compact())

    def _set_value(self, name, value):
        self._values[name] = value

    def _join_intervals(self, name):
        self._intervals.add(name)

        for c in self._children:
            c._join_intervals(name)

        if name in self._values:
            self._values[name] = Interval.join(self._values[name])

    def unique_values(self):
        if self._unique_values is None:
            u = self._unique_values = {}

            for r in self.flatten():
                for k, v in r.items():
                    u.setdefault(k, set())
                    for x in v:
                        u[k].add(x)

            for i in self._intervals:
                u[i] = Interval.join(u[i])

            for k in u.keys():
                u[k] = tuple(sorted(u[k]))

        return dict(**self._unique_values)

    def flatten(self):
        if self._flatten is None:
            self._flatten = tuple(self._flatten_tree())
        return self._flatten

    def _flatten_tree(self):
        if not self._children:
            yield self._values
        else:
            for c in self._children:
                for t in c._flatten_tree():
                    r = dict(**self._values)
                    r.update(t)
                    yield r

    def to_list(self):
        result = []
        for r in _cleanup(self.flatten()):
            s = {k: sorted(v) for k, v in sorted(r.items())}
            result.append(s)

        return sorted(result, key=lambda a: sorted(a.items()))

    def visit(self, visitor, depth=0):
        visitor(self._values, depth)
        for c in self._children:
            c.visit(visitor, depth + 1)

    def _kwargs_to_request(self, _kwargs=None, **kwargs):
        if _kwargs is not None and not kwargs:
            assert isinstance(_kwargs, dict)
            kwargs = _kwargs

        request = {}
        for k, v in kwargs.items():
            if v is None:
                continue
            if k in self._intervals:
                request[k] = _as_interval(v)
            else:
                request[k] = _as_tuple(v)
        return request

    def count(self, _kwargs=None, **kwargs):
        return self._count(self._kwargs_to_request(_kwargs, **kwargs))

    def _count(self, request):
        if not self._values and not self._children:
            return 0

        ok, matches = self._match(request)
        if not ok:
            return 0

        r = dict(**self._values)
        for name, values in [(n, v) for (n, v) in matches.items() if n in self._values]:
            r[name] = _as_tuple(values)

        count = 1
        for name, values in r.items():
            if name in self._intervals:
                count *= sum(i.count() for i in values)
            else:
                count *= len(values)

        if not self._children:
            return count

        return sum(count * c._count(request) for c in self._children)

    def select(self, **kwargs):
        result = self._select(self._kwargs_to_request(**kwargs))
        if result is None:
            return Tree()
        return result.factorise()

    def _select(self, request):
        ok, matches = self._match(request)
        if not ok:
            return None

        r = dict(**self._values)
        for name, values in [(n, v) for (n, v) in matches.items() if n in self._values]:
            r[name] = _as_tuple(values)
        result = Tree(r, self._intervals)

        if not self._children:
            return result

        cnt = 0
        for c in self._children:
            s = c._select(request)
            if s is not None:
                cnt += 1
                result._add_child(s)

        if cnt == 0:
            return None

        return result

    def missing(self, **kwargs):
        request = self._kwargs_to_request(**kwargs)
        user = {_to_hashable(x) for x in self._iterate_request(request)}
        tree = {_to_hashable(x) for x in self.iterate(True)}

        s = [_from_hashable(x) for x in user.difference(user.intersection(tree))]

        return _factorise(s, intervals=self._intervals)

    def _match(self, request):
        matches = {}
        for name, values in [(n, v) for (n, v) in request.items() if n in self._values]:
            if name in self._intervals:
                common = Interval.intersection(values, self._values[name])
            else:
                common = set(values).intersection(set(self._values[name]))

            if len(common) == 0:
                return False, None

            if False:  # If we want an exact match
                if len(common) != len(values):
                    return False, None

            matches[name] = common

        return True, matches

    def iterate(self, expand=False):
        for r in self._flatten_tree():
            if expand:
                yield from self._iterate_request(r)
            else:
                yield r

    def _iterate_request(self, r):
        for name in self._intervals:
            r[name] = Interval.expand(r[name])
        yield from (dict(zip(r.keys(), x)) for x in itertools.product(*r.values()))

    def compact(self):
        if not self._values and len(self._children) == 1:
            return self._children[0]
        return self

    def factorise(self):
        return _factorise(list(self._flatten_tree()), intervals=self._intervals)

    def as_mars(self, verb="retrieve", extra=None):
        result = []
        for r in self.flatten():
            req = [verb]
            if extra is not None:
                req.append(extra)
            for k, v in r.items():
                v = [str(_) for _ in v]
                req.append(f"{k}={'/'.join(v)}")
            result.append(",".join(req))
        return "\n".join(result)

    def as_mars_list(self):
        text = []
        indent = {}
        order = {}

        def V(request, depth):
            if not request:
                return

            if depth not in indent:
                indent[depth] = len(indent)

            text.append(" " * indent[depth])

            for k in sorted(request.keys()):
                if k not in order:
                    order[k] = len(order)

            sep = ""
            for k, v in sorted(request.items(), key=lambda x: order[x[0]]):
                text.append(sep)
                text.append(k)
                text.append("=")

                if isinstance(v[0], Interval):
                    v = [str(x) for x in v]

                if len(v) == 1:
                    text.append(v[0])
                else:
                    start, end = self._to_date_interval(k, v)
                    if start is not None:
                        text.append(f"{start}/to/{end}")
                    else:
                        text.append("/".join([str(_) for _ in sorted(v)]))
                sep = ","
            text.append("\n")

        self.visit(V)

        return "".join(str(x) for x in text)

    def _to_date_interval(self, k, v):
        class ReturnNoneNone(Exception):
            pass

        def parse_date(d):
            try:
                return datetime.datetime.strptime(d, "%Y%m%d")
            except:  # noqa: E722
                raise ReturnNoneNone()

        try:
            if k != "date":
                raise ReturnNoneNone()

            if len(k) < 3:
                raise ReturnNoneNone()

            start = parse_date(str(v[0]))
            step = parse_date(str(v[1])) - start

            if step != datetime.timedelta(days=1):
                raise ReturnNoneNone()

            for i in range(2, len(v)):
                current = parse_date(str(v[i]))
                previous = parse_date(str(v[i - 1]))
                if current - previous != step:
                    print(int(v[i]))
                    print(
                        f"expecting {previous + step} after {previous}, found {current}"
                    )
                    raise ReturnNoneNone()
            return str(v[0]), str(v[-1])

        except ReturnNoneNone:
            return None, None

    def tree(self):
        text = []
        indent = {}
        order = {}

        def V(request, depth):
            if not request:
                return

            if depth not in indent:
                indent[depth] = len(indent) * 3

            text.append(" " * indent[depth])

            for k in sorted(request.keys()):
                if k not in order:
                    order[k] = len(order)

            sep = ""
            for k, v in sorted(request.items(), key=lambda x: order[x[0]]):
                text.append(sep)
                text.append(k)
                text.append("=")

                if isinstance(v[0], Interval):
                    v = [str(x) for x in v]

                if len(v) == 1:
                    text.append(v[0])
                else:
                    start, end = self._to_date_interval(k, v)
                    if start is not None:
                        text.append(f"{start}/to/{end}")
                    else:
                        text.append("[")
                        text.append(", ".join([str(_) for _ in sorted(v)]))
                        text.append("]")
                sep = ", "
            text.append("\n")

        self.visit(V)

        return "".join(str(x) for x in text)


class Column(object):
    """Just what is says on the tin, a column of values."""

    def __init__(self, title, values):
        self.title = title
        self.values = values
        self.prio = 0
        self.diff = -1

    def __lt__(self, other):
        return (self.prio, self.diff, self.title) < (
            other.prio,
            other.diff,
            other.title,
        )

    def value(self, i):
        return self.values[i]

    def set_value(self, i, v):
        self.values[i] = v

    def compute_differences(self, idx):
        """
        Number of unique values in this column for the requested
        row indexes.

        @param idx list of row indexes
        """
        x = [self.values[i] for i in idx]
        try:
            self.diff = len(set(x))
        except Exception:
            print(type(x))
            print(x[:10])
            raise

    def __repr__(self):
        return "Column(%s,%s,%s,%s)" % (self.title, self.values, self.prio, self.diff)


class Table(object):
    def __init__(self, other=None, a=None, b=None):
        self.tree = Tree()

        if other is not None:
            self.depth = other.depth + 1
            self.cols = copy(other.cols)
            self.colidx = copy(other.colidx)
            self.rowidx = other.rowidx[a:b]
        else:
            self.depth = 0
            self.cols = []
            self.colidx = []
            self.rowidx = []

    def get_elem(self, c, r):
        return self.cols[self.colidx[c]].value(self.rowidx[r])

    def set_elem(self, c, r, v):
        return self.cols[self.colidx[c]].set_value(self.rowidx[r], v)

    def __repr__(self):
        return repr(
            [[self.cols[col].value(row) for row in self.rowidx] for col in self.colidx]
        )

    def column(self, s, col):
        self.cols.append(Column(s, col))
        self.colidx.append(len(self.colidx))

        if len(col) > len(self.rowidx):
            self.rowidx = [i for i in range(len(col))]

    def one_less(self, r, n):
        return [self.get_elem(i, r) for i in range(len(self.colidx)) if i != n]

    def factorise1(self):
        self.pop_singles()
        self.sort_rows()

        for i in range(len(self.colidx)):
            self.factorise2(len(self.colidx) - i - 1)

    def factorise2(self, n):
        remap = defaultdict(list)
        gone = []

        for i in range(len(self.rowidx)):
            v = self.one_less(i, n)
            s = remap[_as_tuple(v)]
            if len(s) != 0:
                gone.append(i)
            elem = self.get_elem(n, i)
            if elem not in s:
                s.append(elem)

        for g in reversed(gone):
            del self.rowidx[g]

        for i in range(len(self.rowidx)):
            v = self.one_less(i, n)
            s = remap[_as_tuple(v)]
            self.set_elem(n, i, _as_tuple(s))

    def sort_columns(self):
        """
        Sort the columns on the number of unique values (this column.diff).
        """
        for idx in self.colidx:
            self.cols[idx].compute_differences(self.rowidx)

        self.colidx.sort(key=lambda a: self.cols[a])

    def compare_values(self, sa, sb):
        if isinstance(sa, tuple) and isinstance(sb, tuple):
            for a, b in zip(sa, sb):
                n = self.compare_values(a, b)
                if n != 0:
                    return n
            return 0

        if type(sa) is type(sb):
            if sa < sb:
                return -1

            if sa > sb:
                return 1

            return 0

        return self.compare_values(str(type(sa)), str(type(sb)))

    def compare_rows(self, a, b):
        for idx in self.colidx:
            sa = self.cols[idx].value(a)
            sb = self.cols[idx].value(b)

            if sa is None and sb is None:
                continue

            n = self.compare_values(sa, sb)

            if n != 0:
                return n

        return 0

    def sort_rows(self):
        self.rowidx.sort(key=cmp_to_key(self.compare_rows))

    def pop_singles(self):
        """
        Take the column with just one unique value and add them to the
        tree. Delete their index from the list of column indexes.
        """
        self.sort_columns()
        ok = False
        while len(self.colidx) > 0 and self.cols[self.colidx[0]].diff == 1:
            s = _as_tuple(self.get_elem(0, 0))
            self.tree._set_value(self.cols[self.colidx[0]].title, s)
            del self.colidx[0]
            ok = True

        return ok

    def split(self):
        if len(self.rowidx) < 2 or len(self.colidx) < 2:
            return

        self.sort_columns()
        self.sort_rows()

        prev = self.get_elem(0, 0)
        j = 0

        for i in range(1, len(self.rowidx)):
            e = self.get_elem(0, i)
            if prev != e:
                table = Table(self, j, i)
                self.tree._add_child(table.process())
                j = i
                prev = e

        if j > 0:
            table = Table(self, j, len(self.rowidx))
            self.tree._add_child(table.process())
            self.rowidx = []

    def process(self):
        self.factorise1()
        self.pop_singles()
        self.split()
        return self.tree


def _scan(r, cols, name, rest):
    """Generate all possible combinations of values. Each set of values is
    stored in a list and each value is repeated as many times so that if taking
    one row in all value lists, I will get one unique combination of values
    between all input keys.

    @param r    request as a dict
    @param cols actual result of the _scan
    @param name current request key we are processing
    @param rest remaining request keys to process
    """
    n = 0
    # print("Scan", name)
    c = cols[name]
    for value in r.get(name, ["-"]):
        m = 1
        if rest:
            m = _scan(r, cols, rest[0], rest[1:])
        for _ in range(m):
            c.append(value)
        n += m

    return n


def _as_requests(r):
    s = {}
    for k, v in r.items():
        if not isinstance(v, (tuple, list)):
            s[k] = [v]
        else:
            s[k] = v

    return s


def factorise(req, *, intervals=None):
    # Make a copy so we don't modify the original
    safe = [dict(**r) for r in req]
    return _factorise(safe, intervals=intervals)


def _factorise(req, intervals=None):
    if intervals is None:
        intervals = []

    if not isinstance(intervals, (list, tuple, set)):
        intervals = [intervals]

    if intervals:
        for r in req:
            for i in intervals:
                if i in r:
                    r[i] = _as_interval(r[i])

    for i in intervals:
        # Collect all dates
        dates = set()
        for r in req:
            for interval in r.get(i, []):
                dates.add(interval.start)
                dates.add(interval.end)

        # Split intervals according to collected dates
        for r in req:
            if i in r:
                splits = []
                for interval in r.get(i, []):
                    splits.extend(interval.split(dates))
                r[i] = splits

    req = [_as_requests(r) for r in req]

    names = list({name for r in req for name in r.keys()})

    cols = defaultdict(list)
    if names:
        for r in req:
            _scan(r, cols, names[0], names[1:])

    table = Table()
    for n, c in cols.items():
        table.column(n, c)

    tree = table.process()

    for i in intervals:
        tree._join_intervals(i)

    return tree.compact()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from . import Reader
from .csv import CSVReader, is_csv


def is_text(path, prob_lines=1000, probe_size=4096):
    try:
        with open(path, "rb") as f:
            if 0x0 in f.read(probe_size):
                return False

        with open(path, "r", encoding="utf-8") as f:
            for i, _ in enumerate(f):
                if i > prob_lines:
                    break
        return True
    except UnicodeDecodeError:
        return False


class TextReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)

    def ignore(self):
        # Used by multi-source
        return True

    def mutate(self):
        if is_csv(self.path):
            return CSVReader(self.source, self.path)

        return self


def reader(source, path, magic=None, deeper_check=False):
    if magic is None:  # Bypass check and force
        return TextReader(source, path)

    if deeper_check:
        if is_text(path):
            return TextReader(source, path)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from .aaa import loaded_modules
from .arguments.transformers import ALL
from .core.caching import CACHE as cache
from .core.constants import DATETIME
from .core.initialise import initialise
from .core.settings import SETTINGS as settings
from .datasets import Dataset
from .datasets import get_dataset as dataset  # so the user can do: cml.dataset(...)
from .datasets import load_dataset
from .distributed.dask import start_dask
from .plotting import interactive_map, new_plot, new_table
from .plotting import options as plotting_options
from .plotting import plot_map
from .readers import Reader
from .readers.grib.output import new_grib_output
from .sources import Source
from .sources import get_source as source  # so the user can do: cml.source(...)
from .sources import load_source, load_source_lazily
from .version import __version__
from .wrappers import Wrapper

__all__ = [
    "__version__",
    "ALL",
    "cache",
    "dataset",
    "Dataset",
    "DATETIME",
    "interactive_map",
    "load_dataset",
    "load_source_lazily",
    "load_source",
    "loaded_modules",
    "new_grib_output",
    "new_plot",
    "new_table",
    "plot_map",
    "plotting_options",
    "Reader",
    "settings",
    "source",
    "Source",
    "start_dask",
    "Wrapper",
]


initialise()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import csv
import io
import logging
import mimetypes
import os
import zipfile

from climetlab.wrappers import get_wrapper

from . import Reader

LOG = logging.getLogger(__name__)


class ZipProbe:
    def __init__(self, path, newline=None, encoding=None):
        zip = zipfile.ZipFile(path)
        members = zip.infolist()
        self.f = zip.open(members[0].filename)
        self.encoding = encoding

    def read(self, size):
        bytes = self.f.read(size)
        return bytes.decode(self.encoding)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trace):
        pass


def probe_csv(
    path,
    probe_size=4096,
    compression=None,
    for_is_csv=False,
    minimum_columns=2,
    minimum_rows=2,
):
    OPENS = {
        None: open,
        "zip": ZipProbe,
    }

    try:
        import gzip

        OPENS["gzip"] = gzip.open
    except ImportError:
        pass

    try:
        import bz2

        OPENS["bz2"] = bz2.open
    except ImportError:
        pass

    try:
        import lzma

        OPENS["lzma"] = lzma.open
    except ImportError:
        pass

    _open = OPENS[compression]

    try:
        with _open(path, newline="", encoding="utf-8") as f:
            sample = f.read(probe_size)
            sniffer = csv.Sniffer()
            dialect, has_header = sniffer.sniff(sample), sniffer.has_header(sample)

            LOG.debug("dialect = %s", dialect)
            LOG.debug("has_header = %s", has_header)
            if hasattr(dialect, "delimiter"):
                LOG.debug("delimiter = '%s'", dialect.delimiter)

            if for_is_csv:
                # Check that it is not a trivial text file.
                reader = csv.reader(io.StringIO(sample), dialect)
                if has_header:
                    header = next(reader)
                    LOG.debug("for_is_csv header %s", header)
                    if len(header) < minimum_columns:
                        return None, False
                cnt = 0
                length = None
                for row in reader:
                    cnt += 1
                    LOG.debug("for_is_csv row %s %s", cnt, row)
                    if length is None:
                        length = len(row)
                    if length != len(row):
                        return None, False
                    if cnt >= minimum_rows:
                        break

                if cnt < minimum_rows:
                    return None, False

            return dialect, has_header

    except UnicodeDecodeError:
        return None, False

    except csv.Error:
        return None, False


def is_csv(path, probe_size=4096, compression=None):
    _, extension = os.path.splitext(path)

    if extension in (".xml",):
        return False

    dialect, _ = probe_csv(path, probe_size, compression, for_is_csv=True)
    return dialect is not None


class CSVReader(Reader):
    def __init__(self, source, path, compression=None):
        super().__init__(source, path)
        self.compression = compression
        self.dialect, self.has_header = probe_csv(path, compression=compression)

    def to_pandas(self, **kwargs):
        import pandas

        pandas_read_csv_kwargs = kwargs.get("pandas_read_csv_kwargs", {})
        if self.compression is not None:
            pandas_read_csv_kwargs = dict(**pandas_read_csv_kwargs)
            pandas_read_csv_kwargs["compression"] = self.compression

        LOG.debug("pandas.read_csv(%s,%s)", self.path, pandas_read_csv_kwargs)
        return pandas.read_csv(self.path, **pandas_read_csv_kwargs)

    def to_tfdataset(self, batch_size=1000, **kwargs):
        # See https://www.tensorflow.org/api_docs/python/tf/data/experimental/make_csv_dataset
        import tensorflow as tf

        # TODO: uncompress in cache
        assert self.compression in (None, "gzip")

        options = {}
        if self.dialect is not None:
            COMPRESSIONS = {
                None: None,
                "gzip": "GZIP",
                "????": "ZLIB",  # Which one is that? zip?
            }
            options = dict(
                # column_names=None,
                # column_defaults=None,
                # label_name=None,
                # select_columns=None,
                # # Do not trust the delimiter found by csv sniffer,
                # # as it is wrong for files with only one column.
                # field_delim=self.dialect.delimiter,
                use_quote_delim=self.dialect.doublequote,  # Not sure about that one
                # na_value="",
                header=self.has_header,
                compression_type=COMPRESSIONS[self.compression],
                # ignore_errors=False,
            )

        return tf.data.experimental.make_csv_dataset(
            self.path,  # Can be a pattern
            batch_size,
            **options,
        )

    @classmethod
    def to_tfdataset_multi(cls, paths, batch_size=1000, **kwargs):
        # See https://www.tensorflow.org/api_docs/python/tf/data/experimental/make_csv_dataset
        import tensorflow as tf

        options = {}
        # TODO: read dialect(s) in options

        return tf.data.experimental.make_csv_dataset(
            paths,
            batch_size,
            **options,
        )

    def plot_map(self, backend):
        get_wrapper(self.to_pandas()).plot_map(backend)

    def to_metview(self):
        from climetlab.metview import mv_read_table

        assert self.compression is None
        # TODO: use dialect
        return mv_read_table(table_filename=self.path)


def reader(source, path, magic, deeper_check, fwf=False):
    kind, compression = mimetypes.guess_type(path)

    if kind in [
        "text/csv",
        "application/vnd.ms-excel",  # See: https://github.com/ecmwf/climetlab/issues/48
    ]:
        return CSVReader(source, path, compression=compression)

    if deeper_check and False:
        if is_csv(path):
            return CSVReader(source, path)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging
import os
from urllib.parse import urljoin

import requests
from multiurl import robust

from climetlab.core.caching import cache_file
from climetlab.readers.grib.index import FieldSetInFiles
from climetlab.utils import progress_bar

LOG = logging.getLogger(__name__)


class FieldsetInFilesWithDBIndex(FieldSetInFiles):
    def __init__(self, db, **kwargs):
        """Should not be instanciated directly.
        The public API are the constructors "_from*()" class methods.
        """

        self.db = db

        # self._cache is a tuple : (first, length, result). It holds one chunk of the db.
        # The third element (result) is a list of size length.
        self._cache = None
        self._dict_cache = None

        super().__init__(**kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}({self.db})"

    @property
    def availability_path(self):
        dirpath = os.path.dirname(self.db.db_path)
        return os.path.join(dirpath, "availability.pickle")

    @classmethod
    def from_iterator(
        cls,
        iterator,
        cache_metadata,
        selection=None,
        order_by=None,
        **kwargs,
    ):
        def load(target, *args):
            LOG.debug(f"Building db in {target}")
            db = cls.DBCLASS(target)
            db.load_iterator(iterator)

        db_name = cache_file(
            "grib-index",
            load,
            cache_metadata,
            hash_extra=cls.DBCLASS.VERSION,
            extension=cls.DBCLASS.EXTENSION,
        )

        db = cls.DBCLASS(db_name)

        new = cls(db=db, **kwargs)
        new = new.sel(selection)
        new = new.order_by(order_by)
        return new

    @classmethod
    def from_url(cls, url, patch_entry=None, **kwargs):
        """Create a database from a given url"""

        # If this is a file, open it without download
        if os.path.exists(url):
            return cls.from_file(path=url, **kwargs)
        if url.startswith("file://") and os.path.exists(url[7:]):
            return cls.from_file(path=url[7:], **kwargs)

        # Request to download the data
        r = robust(requests.get)(url, stream=True)
        r.raise_for_status()
        try:
            size = int(r.headers.get("Content-Length"))
        except Exception:
            size = None

        if patch_entry is None:

            def absolute_url(entry):  # closure on "url"
                entry["_path"] = urljoin(url, entry.pop("_path"))
                return entry

            patch_entry = absolute_url

        def progress(iterable):
            pbar = progress_bar(iterable=iterable, desc="Downloading index", total=size)
            for line in pbar:
                yield line
                pbar.update(len(line) + 1)

        def parse_lines(lines):
            for line in lines:
                entry = json.loads(line)
                entry = patch_entry(entry)
                yield entry

        iterator = r.iter_lines()
        iterator = progress(iterator)
        iterator = parse_lines(iterator)

        return cls.from_iterator(
            iterator=iterator,
            cache_metadata={"url": url},
            **kwargs,
        )

    @classmethod
    def from_file(cls, path, **kwargs):
        directory = os.path.dirname(path)
        size = os.path.getsize(path)

        def progress(lines):
            pbar = progress_bar(iterable=lines, desc="Parsing index", total=size)
            for line in pbar:
                yield line
                pbar.update(len(line) + 1)

        def parse_lines(lines):
            for line in lines:
                entry = json.loads(line)
                if not os.path.isabs(entry["_path"]):
                    entry["_path"] = os.path.join(directory, entry["_path"])
                yield entry

        iterator = open(path)
        iterator = progress(iterator)
        iterator = parse_lines(iterator)

        return cls.from_iterator(
            iterator=iterator,
            cache_metadata={"path": path},
            **kwargs,
        )

    @classmethod
    def from_existing_db(cls, db_path, **kwargs):
        assert os.path.exists(db_path)
        return cls(cls.DBCLASS(db_path), **kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.wrappers import Wrapper

LATLON = (
    ("lat@hdr", "lon@hdr"),
    ("lat", "lon"),
    ("latitude", "longitude"),
    ("LAT", "LON"),
    ("LATITUDE", "LONGITUDE"),
)


class PandasSeriesWrapper(Wrapper):
    def __init__(self, series, **kwargs):
        self.series = series

    def to_datetime_list(self):
        return sorted(set(self.series.values))


class PandasFrameWrapper(Wrapper):
    def __init__(self, frame, **kwargs):
        self.frame = frame

        self.lat = "cannot-find-latitude-column"
        self.lon = "cannot-find-longitude-column"

        self.time = "time"
        if "time" not in self.frame:
            self.time = "date"

        for lat, lon in LATLON:
            if lat in self.frame:
                self.lat, self.lon = lat, lon
                break

    def plot_map(self, backend):
        column = backend.option("column", self.lat)

        north, west, south, east = self.bounding_box()

        backend.bounding_box(
            north=north,
            south=south,
            west=west,
            east=east,
        )

        backend.plot_pandas(self.frame, self.lat, self.lon, column)

    def bounding_box(self):
        north = self.frame[self.lat].max()
        south = self.frame[self.lat].min()

        lons1 = self.frame[self.lon]
        east1 = lons1.max()
        west1 = lons1.min()

        lons2 = self.frame[self.lon] % 360
        east2 = lons2.max()
        west2 = lons2.min()

        if abs(east1 - west1) <= abs(east2 - west2):
            east, west = east1, west1
        else:
            east, west = east2, west2

        return [north, west, south, east]

    def to_datetime_list(self):
        return sorted(set(self.frame[self.time].values))

    def to_bounding_box(self):
        return self.bounding_box()


class DatetimeIndexWrapper(Wrapper):
    def __init__(self, index, **kwargs):
        self.index = index

    def to_datetime_list(self):
        return [d.to_pydatetime() for d in self.index]


def wrapper(data, *args, **kwargs):
    import pandas as pd

    if isinstance(data, pd.DatetimeIndex):
        return DatetimeIndexWrapper(data, *args, **kwargs)

    if isinstance(data, pd.DataFrame):
        return PandasFrameWrapper(data, *args, **kwargs)

    if isinstance(data, pd.Series):
        return PandasSeriesWrapper(data, *args, **kwargs)

    return None



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

from climetlab import load_source

LOG = logging.getLogger(__name__)


class LazySourceError(Exception):
    pass


class LazySource:
    def __init__(self, name, *args, **kwargs):
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self._source = None
        self._exception = None

    @property
    def source(self):
        if self._source is None:
            try:
                self._source = load_source(
                    self.name,
                    lazily=False,
                    *self.args,
                    **self.kwargs,
                )
            # except AttributeError as e:
            #     self._exception = e
            #     raise LazySource(e)
            except Exception as e:
                self._exception = e
                raise

        return self._source

    def __getitem__(self, name):
        return self.source[name]

    def __iter__(self):
        return iter(self.source)

    def __len__(self):
        return len(self.source)

    def __getattr__(self, name):
        if self._exception is not None:
            raise self._exception(name)
        assert name != "source"
        return getattr(self.source, name)

    def __call__(self, **kwargs):
        assert self._source is None
        self.kwargs.update(kwargs)
        return self.source



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from climetlab.sources.file import FileSource

LOG = logging.getLogger(__name__)


class Metview(FileSource):
    def __init__(self, metview_object):
        super().__init__(path=metview_object.url())


source = Metview



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
import os

LOG = logging.getLogger(__name__)

global _MIRRORS


class MirrorConnection:
    def __init__(self, mirror, source):
        self.mirror = mirror
        self.source = source

    def get_file(self, create, args):
        if self.resource():
            LOG.debug(
                f"Found a copy of {self.source} in mirror {self.mirror}: {self.resource()}."
            )
            return self.resource()
        if not self.mirror._prefetch:
            LOG.debug(f"No copy of {self.source} into {self.mirror}: prefetch=False.")
            return None
        LOG.info(f"Building mirror for {self.source} in mirror {self.mirror}.")
        return self.create_copy(create, args)

    def resource(self):
        LOG.info(f"Not implemented. {self.source} not in mirror {self.mirror}.")
        return None

    def create_copy(self, create, args):
        LOG.info(
            f"Not implemented. Not creating anything for {self.source} in mirror {self.mirror}."
        )
        return None


class BaseMirror:
    _prefetch = False

    def __enter__(self):
        self.activate(prefetch=self._prefetch)
        return self

    def __exit__(self, *args, **kwargs):
        self.deactivate()

    def prefetch(self):
        self._prefetch = True
        return self

    def activate(self, prefetch=False):
        self._prefetch = prefetch
        global _MIRRORS
        _MIRRORS.append(self)
        if len(_MIRRORS) > 1:
            LOG.warn("Using multiple mirrors is not supported.")

    def deactivate(self):
        self._prefetch = False
        global _MIRRORS
        _MIRRORS.remove(self)

    # convenience method for testing purposes
    def contains(self, source):
        return source.connect_to_mirror(self).resource() is not None

    def connection_for_url(self, source, *args, **kwargs):
        return None

    def connection_for_eod(self, source, *args, **kwargs):
        return None


class Mirror(BaseMirror):
    # TODO: build mirror from json config
    pass


def build_mirror_from_env_var():
    from .directory_mirror import DirectoryMirror

    env_var = os.environ.get("CLIMETLAB_MIRROR")
    if not env_var:
        return None

    if " " in env_var:
        # export CLIMETLAB_MIRROR='https://object-store.os-api.cci1.ecmwf.int file:///data/mirror/https/object-store.os-api.cci1.ecmwf.int' # noqa
        LOG.warning(
            "Deprecation warning: this use of CLIMETLAB_MIRROR environment variable"
            " to define a mirror will be deprecated."
        )
        origin_prefix, path = env_var.split(" ")
        # assert is_url(origin_prefix), "Cannot parse CLIMETLAB_MIRROR={env_var}."
        # assert is_url(path), "Cannot parse CLIMETLAB_MIRROR={env_var}."
        return DirectoryMirror(path=path, origin_prefix=origin_prefix)

    return DirectoryMirror(path=env_var)


def get_active_mirrors():
    global _MIRRORS
    return _MIRRORS


def _reset_mirrors(use_env_var):
    global _MIRRORS
    _MIRRORS = []

    if use_env_var:
        mirror = build_mirror_from_env_var()
        if mirror:
            _MIRRORS.append(mirror)


_reset_mirrors(use_env_var=True)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging
import os
import sys

from climetlab.indexing.database.json import JsonStdoutDatabase
from climetlab.readers.grib.parsing import (
    GribIndexingDirectoryParserIterator,
    _index_url,
)

from .tools import parse_args

LOG = logging.getLogger(__name__)


class GribCmd:
    @parse_args(
        urls=(
            None,
            dict(metavar="URLS", type=str, nargs="+", help="List of urls to index."),
        ),
        baseurl=dict(
            metavar="BASEURL",
            type=str,
            help="Base url to use as a prefix to happen on each URLS to build actual urls.",
        ),
    )
    def do_index_urls(self, args):
        """Create json index files for remote Grib urls.
        If the option --baseurl is provided, the given url are relative to the BASEURL.
        This allows creating an index for multiple gribs."""

        if args.baseurl:

            def resolve_url(u):
                if u.startswith(args.baseurl):
                    # already absolute url
                    return u
                return f"{args.baseurl}/{u}"

        else:

            def resolve_url(url):
                return url

        stdout = JsonStdoutDatabase()
        for u in args.urls:
            url = resolve_url(u)
            for e in _index_url(url):
                e["_path"] = u
                stdout.load_iterator([e])

    @parse_args(
        url=(None, dict(metavar="URL", type=str, help="url to index")),
        # output=dict(type=str, help="Output filename"),
    )
    def do_index_url(self, args):
        """Create json index files for remote Grib url."""
        stdout = JsonStdoutDatabase()
        for e in _index_url(args.url):
            assert "_path" not in e
            stdout.load_iterator([e])

    @parse_args(
        directory=(None, dict(help="Directory containing the GRIB files to index.")),
        # pattern=dict(help="Files to index (patterns).", nargs="*"),
        no_follow_links=dict(action="store_true", help="Do not follow symlinks."),
        relative_paths=dict(
            action="store_true",
            help=(
                "Use relative paths. Default is to use relative paths, "
                "except when a custom location is provided for the index location. "
                "(when the argument --output is used, default for --relative-path is False)"
            ),
        ),
        output=(
            "--output",
            dict(
                help="Custom location of the database file, will write absolute filenames in the database."
            ),
        ),
    )
    def do_index_directory(self, args):
        """Index a directory containing GRIB files."""

        if sys.platform == "win32":
            print("Not supported on windows")
            return

        directory = args.directory
        db_path = args.output
        force_relative_paths_on = args.relative_paths

        followlinks = True
        if args.no_follow_links:
            followlinks = False

        assert os.path.isdir(directory), directory

        from climetlab.sources.indexed_directory import IndexedDirectorySource

        if force_relative_paths_on:
            relative_paths = True
        else:
            if db_path is None:
                relative_paths = True
            else:
                LOG.warning(
                    (
                        "Non-default location for the index file (--output),"
                        " using absolute path in the index (i.e. setting option --relative-paths to False)."
                    )
                )
                relative_paths = False

        if db_path is None:
            db_path = os.path.join(directory, IndexedDirectorySource.DEFAULT_DB_FILE)

        parser = GribIndexingDirectoryParserIterator(
            directory,
            db_path=db_path,
            relative_paths=relative_paths,
            followlinks=followlinks,
            with_statistics=True,
        )
        parser.load_database()

    @parse_args(filename=(None, dict(help="Database filename.")))
    def do_dump_index(self, args):
        db_path = args.filename

        from climetlab.indexing.database.sql import SqlDatabase
        from climetlab.sources.indexed_directory import IndexedDirectorySource

        if os.path.isdir(db_path):
            db_path = os.path.join(db_path, IndexedDirectorySource.DEFAULT_DB_FILE)

        db = SqlDatabase(db_path=db_path)
        for d in db.lookup_dicts():
            print(json.dumps(d))



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import io
import os.path
import sys

import setuptools


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return io.open(file_path, encoding="utf-8").read()


version = None
lines = read("climetlab/version").split("\n")
if lines:
    version = lines[0]


assert version

numpy = "numpy"
pandas = "pandas"
dask = "dask"


if sys.version_info < (3, 8):
    if not os.environ.get("CLIMETLAB_RUN_UNSUPPORTED_PYTHON_VERSION"):
        raise Exception(
            "Python version "
            + str(sys.version_info)
            + " is not supported. Python 3.8 is required."
        )

install_requires = []
if sys.version_info < (3, 7):
    install_requires += [
        "numpy<1.20",
        "pandas==1.1.5",
        "dataclasses",  # Needed by dask
        "xarray",
    ]
else:
    install_requires += ["numpy", "pandas", "xarray>=0.19.0"]

install_requires += [
    # need to install  to avoid conflict between aiohttp (dependency of s3fs) and requests (cdsapi)
    # "chardet>=3.0,<4.0",
    # "aiohttp>=3.7.2",
    # --
    "requests",
    # "zarr",
    # "s3fs",
    "dask",
    "netcdf4",
    "cfgrib>=0.9.10.1",
    "cdsapi",
    "ecmwf-api-client>=1.6.1",
    "multiurl>=0.1.0",
    "ecmwf-opendata>=0.1.2",
    "tqdm",
    "eccodes>=1.3.0",
    "magics>=1.5.6",
    "pdbufr",
    "pyodc",
    "toolz",
    "filelock",
    "pyyaml",
    "markdown",
    "termcolor",
    "entrypoints",
    "branca",
    "imageio",
    "numpngw",
    "lru-dict",
]
if not os.environ.get("CLIMETLAB_DO_NOT_INSTALL_ECMWFLIBS"):
    install_requires.append("ecmwflibs")

extras_require = {
    "interactive": [
        "skinnywms",
        "folium>=0.12.1",
    ],
    "tensorflow": [
        "tensorflow",
    ],
    "zarr": [
        "zarr",
        "s3fs",
    ],
}


full = []
for k, v in extras_require.items():
    full += v
full += install_requires

extras_require["full"] = full


setuptools.setup(
    name="climetlab",
    version=version,
    author="ECMWF",
    author_email="software.support@ecmwf.int",
    license="Apache 2.0",
    url="https://github.com/ecmwf/climetlab",
    description="Handling of climate/meteorological data",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require,
    zip_safe=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
    ],
    tests_require=[
        "pytest",
        "nbconvert",
        "jupyter",
        "pytest-cov",
        "climetlab-demo-dataset",
        "climetlab-demo-source",
    ],
    test_suite="tests",
    entry_points={"console_scripts": ["climetlab=climetlab.scripts:main"]},
)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.core import Base
from climetlab.core.index import Index, MaskIndex, MultiIndex


class Field(Base):
    pass


class FieldSet(Index):
    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return MaskFieldSet(*args, **kwargs)

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(_, FieldSet) for _ in sources)
        return MultiFieldSet(sources)


class MaskFieldSet(FieldSet, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)


class MultiFieldSet(FieldSet, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)


class FieldArray(FieldSet):
    def __init__(self, fields=None):
        self.fields = fields if fields is not None else []

    def append(self, field):
        self.fields.append(field)

    def _getitem(self, n):
        return self.fields[n]

    def __len__(self):
        return len(self.fields)

    def __repr__(self) -> str:
        return f"FieldArray({len(self.fields)})"



#!/usr/bin/env python3
# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


class YAFilter:
    def __init__(self, source):
        self.init_source = source
        self._source = None
        self._func = None

    def apply_to_source(self, source):
        # To be overridden by subclasses
        return source

    def apply_to_func(self, func):
        # To be overridden by subclasses
        return func

    @property
    def source(self):
        if self._source is None:
            self._source = self.apply_to_source(self.init_source)
        return self._source

    @property
    def func(self):
        if self._func is None:

            def fn(i):
                return self.source[i].to_numpy()

            self._func = self.apply_to_func(fn)
        return self._func

    def __call__(self, i):
        return self.func(i)


class NormaliseFilter(YAFilter):
    def __init__(self, source, a, b):
        super().__init__(source)
        self.a = a
        self.b = b

    def apply_to_func(self, func):
        def fn(i):
            arr = func(i)
            arr = self.a * arr + self.b
            return arr

        return fn


class NormaliseMeanStdFilter(YAFilter):
    def __init__(self, source):
        a, b = compute_mean_std_from_source(source)  # noqa: TODO
        super().__init__(source, a, b)


class NormaliseMinMaxFilter(YAFilter):
    def __init__(self, source):
        a, b = compute_min_max_from_source(source)  # noqa: TODO
        super().__init__(source, a, b)


class OffsetFilter(YAFilter):
    def __init__(self, offset):
        self.offset = offset

    def apply_to_source(self, source):
        # Create a Wrapper class?
        # decrease the len
        return source

    def apply_to_func(self, func):
        def fn(i):
            return func(i + self.offset)

        return fn



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from . import Reader

LOG = logging.getLogger(__name__)


class TfRecordReader(Reader):
    _tfrecord = None
    _len = None

    @property
    def tfrecord(self):
        return type(self).to_tfdataset_multi([self.path])

    def to_tfdataset(self, **kwargs):
        return self.tfrecord

    @classmethod
    def to_tfdataset_multi(cls, paths, **kwargs):
        import tensorflow as tf

        files_ds = tf.data.Dataset.list_files(paths)
        options = tf.data.Options()
        # options.experimental_deterministic = False
        files_ds = files_ds.with_options(options)
        return tf.data.TFRecordDataset(
            files_ds,
            num_parallel_reads=tf.data.experimental.AUTOTUNE,
        )

    def __len__(self):
        if self._len is None:
            record = self.tfrecord
            try:
                self._len = len(record)
            except TypeError:
                self._len = 0
                for _ in record:
                    self._len += 1
        return self._len


def reader(source, path, magic=None, deeper_check=False):
    if magic is None or path.endswith(".tfrecord"):
        return TfRecordReader(source, path)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import inspect
import logging

LOG = logging.getLogger(__name__)


class ArgsKwargs:
    def __init__(self, args, kwargs, func=None):
        assert isinstance(kwargs, dict)
        assert isinstance(args, (list, tuple))
        args = list(args)

        self.args = args
        self.kwargs = kwargs
        self.func = func

        self.positionals_only = []
        self.defaults = {}

    def add_default_values_and_kwargs(self):
        new_args = []
        new_kwargs = {}

        sig = inspect.signature(self.func)
        bnd = sig.bind(*self.args, **self.kwargs)
        parameters_names = list(sig.parameters)

        bnd.apply_defaults()

        new_kwargs.update(bnd.kwargs)

        # TODO: delete this (sic!)
        #
        # if parameters_names[0] == "self":
        #     # func must be method. Store first argument and skip it latter
        #     LOG.debug('Skipping first parameter because it is called "self"')
        #     new_args = new_args + [self.args.pop(0)]
        #     parameters_names.pop(0)

        for name in parameters_names:
            param = sig.parameters[name]

            if param.default != inspect.Parameter.empty:
                self.defaults[name] = param.default

            if param.kind is param.VAR_POSITIONAL:  # param is *args
                new_args = new_args + self.args
                continue

            if param.kind is param.VAR_KEYWORD:  # param is **kwargs
                new_kwargs.update(bnd.arguments[name])
                continue

            if param.kind is param.POSITIONAL_ONLY:
                # new_args = new_args + [bnd.arguments[name]]
                # continue
                self.positionals_only.append(name)

            new_kwargs[name] = bnd.arguments[name]

        LOG.debug("Fixed input arguments args=%s, kwargs=%s", new_args, new_kwargs)

        self.args = new_args
        self.kwargs = new_kwargs

    def ensure_positionals_only(self):
        """Move positional arguments from self.kwargs into self.args"""
        for name in self.positionals_only:
            value = self.kwargs.pop(name)
            self.args.append(value)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging
import os
import re
import stat
from getpass import getpass

import markdown

from climetlab.core.ipython import HTML, display, ipython_active

LOG = logging.getLogger(__name__)


# See https://medium.com/analytics-vidhya/the-ultimate-markdown-guide-for-jupyter-notebook-d5e5abf728fd
HTML_MESSAGE = """
<div style='border: 1px solid orange; color: black;
     background-color: rgb(255, 214, 0);
     margin: 0.5em; padding: 0.5em; font-weight: bold;'>
{message}
</div>
"""

HTML_ASK = """
<div style='border: 1px solid gray; color: black;
     background-color: rgb(230, 230, 230);
     margin: 0.2em; padding: 0.2em; font-weight: bold;'>
{message}
</div>
"""

MESSAGE = """
An API key is needed to access this dataset. Please visit
{register_or_sign_in_url} to register or sign-in
then visit {retrieve_api_key_url} to retrieve you API key.
"""


class RegexValidate:
    def __init__(self, pattern):
        self.pattern = pattern

    def __call__(self, value):
        assert re.fullmatch(self.pattern, value), (self.pattern, value)
        return value


class Prompt:
    def __init__(self, owner):
        self.owner = owner

    def ask_user(self):
        self.print_message()

        result = {}

        for p in self.owner.prompts:
            method = getpass if p.get("hidden", False) else input
            value = self.ask(p, method)
            value = value.strip() or p.get("default", "")

            validate = p.get("validate")
            if validate:
                if isinstance(validate, str):
                    validate = RegexValidate(validate)

                value = validate(value)

            result[p["name"]] = value

        return result


class Text(Prompt):
    def print_message(self):
        print(
            MESSAGE.format(
                register_or_sign_in_url=self.owner.register_or_sign_in_url,
                retrieve_api_key_url=self.owner.retrieve_api_key_url,
            )
        )

    def ask(self, p, method):
        message = f"Please enter a value for '{p.get('title')}'"
        if "default" in p:
            message += f" or leave empty for the default value '{p.get('default')}'"
        message += ", then press <ENTER>"
        return method(p.get("title") + ": ").strip()


class Markdown(Prompt):
    def print_message(self):
        message = markdown.markdown(
            MESSAGE.format(
                register_or_sign_in_url=f"<{self.owner.register_or_sign_in_url}>",
                retrieve_api_key_url=f"<{self.owner.retrieve_api_key_url}>",
            )
        )
        # We use Python's markdown instead of IPython's Markdown because
        # jupyter lab/colab/deepnotes all behave differently
        display(HTML(HTML_MESSAGE.format(message=message)))

    def ask(self, p, method):
        message = f"Please enter a value for <span style='color: red;'>{p.get('title')}</span>"
        if "default" in p:
            message += f" or leave empty for the default value <span style='color: red;'>{p.get('default')}</span>"
        message += ", then press *ENTER*"
        if "example" in p:
            message += f" The value should look like  <span style='color: red;'>{p.get('example')}</span>"
        message = markdown.markdown(message)
        display(HTML(HTML_ASK.format(message=message)))
        return method(p.get("title") + ": ").strip()


class APIKeyPrompt:
    def check(self, load=False):
        rcfile = os.path.expanduser(self.rcfile)
        if not os.path.exists(rcfile):
            self.ask_user_and_save()

        if load:
            with open(rcfile) as f:
                return self.load(f)

    def ask_user(self):
        if ipython_active:
            prompt = Markdown(self)
        else:
            prompt = Text(self)

        return self.validate(prompt.ask_user())

    def ask_user_and_save(self):
        input = self.ask_user()

        rcfile = os.path.expanduser(self.rcfile)
        with open(rcfile, "w") as f:
            self.save(input, f)

        LOG.info("API key saved to '%s'", rcfile)

        try:
            os.chmod(rcfile, stat.S_IREAD | stat.S_IWRITE)
        except OSError:
            LOG.exception("Cannot change access to rcfile")

        return input

    def save(self, input, file):
        json.dump(input, file, indent=4)

    def load(self, file):
        return json.load(file)

    def validate(self, input):
        return input



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging
from collections import defaultdict

from climetlab.core.caching import auxiliary_cache_file
from climetlab.utils.bbox import BoundingBox

from .pandas import PandasMixIn
from .pytorch import PytorchMixIn
from .tensorflow import TensorflowMixIn
from .xarray import XarrayMixIn

LOG = logging.getLogger(__name__)


class FieldSetMixin(PandasMixIn, XarrayMixIn, PytorchMixIn, TensorflowMixIn):
    _statistics = None

    def _find_all_coords_dict(self):
        from climetlab.indexing.database import GRIB_KEYS_NAMES

        coords = defaultdict(set)
        for f in self:
            for k in GRIB_KEYS_NAMES:
                v = f.metadata(k)
                if v is None:
                    continue
                coords[k].add(v)

        return {k: list(v) for k, v in coords.items()}

    @property
    def all_coords(self):
        if not self._coords:
            self._coords = self._find_all_coords_dict()
        return self._coords

    @property
    def coords(self):
        # squeeze coords with only 1 value.
        return {k: v for k, v in self.all_coords.items() if len(v) > 1}

    @property
    def first(self):
        return self[0]

    def to_metview(self):
        from climetlab.metview import mv_read

        return mv_read(self.path)

    def to_numpy(self, **kwargs):
        import numpy as np

        return np.array([f.to_numpy(**kwargs) for f in self])

    def plot_map(self, backend):
        return self.first.plot_map(backend)

    # Used by normalisers
    def to_datetime(self):
        times = self.to_datetime_list()
        assert len(times) == 1
        return times[0]

    def to_datetime_list(self):
        # TODO: check if that can be done faster
        result = set()
        for s in self:
            result.add(s.valid_datetime())
        return sorted(result)

    def to_bounding_box(self):
        return BoundingBox.multi_merge([s.to_bounding_box() for s in self])

    def statistics(self):
        import numpy as np

        if self._statistics is not None:
            return self._statistics

        if False:
            cache = auxiliary_cache_file(
                "grib-statistics--",
                self.path,
                content="null",
                extension=".json",
            )

            with open(cache) as f:
                self._statistics = json.load(f)

            if self._statistics is not None:
                return self._statistics

        stdev = None
        average = None
        maximum = None
        minimum = None
        count = 0

        for s in self:
            v = s.values
            if count:
                stdev = np.add(stdev, np.multiply(v, v))
                average = np.add(average, v)
                maximum = np.maximum(maximum, v)
                minimum = np.minimum(minimum, v)
            else:
                stdev = np.multiply(v, v)
                average = v
                maximum = v
                minimum = v

            count += 1

        nans = np.count_nonzero(np.isnan(average))
        assert nans == 0, "Statistics with missing values not yet implemented"

        maximum = np.amax(maximum)
        minimum = np.amin(minimum)
        average = np.mean(average) / count
        stdev = np.sqrt(np.mean(stdev) / count - average * average)

        self._statistics = dict(
            minimum=minimum,
            maximum=maximum,
            average=average,
            stdev=stdev,
            count=count,
        )

        if False:
            with open(cache, "w") as f:
                json.dump(self._statistics, f)

        return self._statistics

    def save(self, filename):
        with open(filename, "wb") as f:
            self.write(f)

    def write(self, f):
        for s in self:
            s.write(f)



import folium
import folium.plugins
from branca.element import MacroElement
from folium.map import Layer
from jinja2 import Template

from climetlab.core.ipython import HTML, guess_which_ipython


class SVGOverlay(Layer):
    _name = "SVGOverlay"
    _template = Template(
        """
        {% macro script(this, kwargs) %}
            var svgElement = document.createElementNS("http://www.w3.org/2000/svg", "svg");
            svgElement.setAttribute('xmlns', "http://www.w3.org/2000/svg");
            svgElement.setAttribute('viewBox', {{ this.viewBox|tojson }});
            svgElement.innerHTML = {{ this.innerHTML|tojson }}
            var {{ this.get_name() }} = L.svgOverlay(
                svgElement,
                {{ this.bounds|tojson }},
                {{ this.options|tojson }}
            ).addTo({{ this._parent.get_name() }});
        {% endmacro %}"""
    )

    def __init__(self, path, bounds, options={}):
        super().__init__()
        self.path = path
        self.bounds = bounds
        self.options = options

        lines = []
        ok = False
        with open(self.path) as f:
            for line in f:
                if line.startswith("viewBox="):
                    self.viewBox = line.split('"')[1]

                if line.startswith('<metadata id="MAGICSmetadata">'):
                    ok = True
                    continue

                if line.startswith("</svg>"):
                    ok = False
                    continue

                if ok:
                    lines.append(line)

        self.innerHTML = "".join(lines)

    def _get_self_bounds(self):
        return self.bounds


class NoScrollZoom(MacroElement):
    _name = "NoScrollZoom"
    _template = Template(
        """
        {% macro header(this,kwargs) %}
        {% endmacro %}
        {% macro html(this,kwargs) %}
        {% endmacro %}
        {% macro script(this,kwargs) %}
        {{ this._parent.get_name() }}.scrollWheelZoom.disable();
        {% endmacro %}
    """
    )


def make_map(path, bbox, **kwargs):
    center = (0, 0)
    zoom = 1
    # fmt = "image/svg+xml"
    # fmt = "image/png"

    if bbox is not None:
        center = (bbox.north + bbox.south) / 2, (bbox.east + bbox.west) / 2
        zoom = 1 / max((bbox.north - bbox.south) / 180, (bbox.east - bbox.west) / 360)
        zoom = (2 * zoom + 88) / 27

    m = folium.Map(
        zoom_start=zoom,
        location=center,
        # tiles=None,
        tiles="https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png",
        attr='Map data &copy; <a href="https://www.openstreetmap.org/copyright">OpenStreetMap</a> contributors',
    )  # tiles=None

    # folium.raster_layers.WmsTileLayer(
    #     url="https://apps.ecmwf.int/wms/?token=public",
    #     layers="background",
    #     fmt=fmt,
    #     # styles="cream_sky_background",
    #     # transparent=True,
    #     overlay=False,
    # ).add_to(m)
    # folium.raster_layers.WmsTileLayer(
    #     url="https://apps.ecmwf.int/wms/?token=public",
    #     layers="background",
    #     fmt=fmt,
    #     styles="black_admin_boundaries",
    #     transparent=True,
    #     # overlay=False,
    # ).add_to(m)
    SVGOverlay(
        path=path,
        # bounds=[[85.051129, -180],[-85.051129, 180]],
        bounds=[[90, -180], [-90, 180]],
        # options=dict(opacity=0.6, autoZIndex=True),
        options=dict(attribution="CliMetLab"),
    ).add_to(m)

    folium.plugins.Fullscreen(force_separate_button=True).add_to(m)
    NoScrollZoom().add_to(m)

    # folium.raster_layers.WmsTileLayer(
    #     url="https://apps.ecmwf.int/wms/?token=public",
    #     layers="foreground",
    #     fmt=fmt,
    #     transparent=True,
    #     styles="medium_res_foreground",
    #     # overlay=False,
    # ).add_to(m)

    # folium.raster_layers.WmsTileLayer(
    #     url="https://apps.ecmwf.int/wms/?token=public",
    #     layers="grid",
    #     fmt=fmt,
    #     transparent=True,
    #     # overlay=False,
    # ).add_to(m)

    if bbox is not None:
        m.fit_bounds([[bbox.south, bbox.east], [bbox.north, bbox.west]])

    html = m._repr_html_()

    if guess_which_ipython()[0] == "deepnote":
        # For deepnote
        html = html.replace("width: 100%;height: 100%", "width: 100%").replace(
            "height: 100.0%;", "height: 609px;"
        )

    return HTML(html)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime
import logging
import os
import sys
import time
from multiprocessing import Process, Queue

from tqdm import tqdm

from climetlab.utils import progress_bar
from climetlab.utils.humanize import plural, seconds

LOG = logging.getLogger(__name__)


def post_process_valid_date(field, h):
    date = h.get("validityDate")
    time = h.get("validityTime")
    field["datetime"] = datetime.datetime(
        date // 10000,
        date % 10000 // 100,
        date % 100,
        time // 100,
        time % 100,
    )
    # Note that we do not create a script here:
    # There is no ".isoformat()". Because the sql database
    # takes care of this conversion back and forth.
    return field


def post_process_parameter_level(field, h):
    param = field.get("param", None)
    if param is None:
        field["param_level"] = None
        return field

    level = field.get("levelist", None)
    if level is None:
        field["param_level"] = param
        return field

    field["param_level"] = f"{param}_{level}"
    return field


def post_process_statistics(field, h):
    values = h.get("values")
    field["mean"] = values.mean()
    field["std"] = values.std()
    field["min"] = values.min()
    field["max"] = values.max()
    field["shape"] = ",".join([str(_) for _ in values.shape])
    return field


def _index_grib_file(
    path,
    with_statistics=False,
    with_valid_date=True,
    with_parameter_level=True,
    position=0,
):
    import eccodes

    from climetlab.readers.grib.codes import CodesHandle

    post_process_mars = []
    if with_valid_date:
        post_process_mars.append(post_process_valid_date)
    if with_parameter_level:
        post_process_mars.append(post_process_parameter_level)
    if with_statistics:
        post_process_mars.append(post_process_statistics)

    def parse_field(h):
        field = h.as_mars()

        if post_process_mars:
            for f in post_process_mars:
                field = f(field, h)

        field["_path"] = path
        field["_offset"] = h.get_long("offset")
        field["_length"] = h.get_long("totalLength")
        field["_param_id"] = h.get_string("paramId")
        field["md5_grid_section"] = h.get("md5GridSection")

        # eccodes.codes_get_string(h, "number") returns "0"
        # when "number" is not in the iterator
        # remove? field["number"] = h.get("number")

        return field

    size = os.path.getsize(path)
    pbar = tqdm(
        desc=f"Parsing {path}",
        total=size,
        unit_scale=True,
        unit_divisor=1024,
        unit="B",
        leave=False,
        # position=TQDM_POSITION,
        position=position,
        dynamic_ncols=True,
    )

    with open(path, "rb") as f:
        old_position = f.tell()
        h = eccodes.codes_grib_new_from_file(f)

        while h:
            yield parse_field(CodesHandle(h, path, offset=old_position))

            position = f.tell()
            pbar.update(position - old_position)
            old_position = position
            h = eccodes.codes_grib_new_from_file(f)

    pbar.close()


def _index_url(url):
    import climetlab as cml

    path = cml.load_source("url", url).path
    # TODO: should use download_and_cache
    # path = download_and_cache(url)
    for entry in _index_grib_file(path):
        del entry["_path"]
        yield entry


def _index_path(path):
    if os.path.isdir(path):
        for root, _, files in os.walk(path):
            for name in files:
                yield from _index_grib_file(os.path.join(root, name))
    else:
        yield from _index_grib_file(path)


class GribIndexingDirectoryParserIterator:
    """This class delays parsing the directory for the list of files
    until the iterator is actually used (calling __iter__)
    """

    def __init__(
        self,
        directory,
        db_path,
        relative_paths,
        extensions=[".grib", ".grib1", ".grib2"],
        followlinks=True,
        verbose=False,
        with_statistics=True,
    ):
        self.db_path = db_path
        self.extensions = set(extensions)
        self.directory = directory
        self.relative_paths = relative_paths
        self.followlinks = followlinks
        self.verbose = verbose
        self.with_statistics = with_statistics

        self._tasks = None

    def _new_db(self):
        from climetlab.indexing.database.sql import SqlDatabase

        return SqlDatabase(self.db_path)

    def worker(self, i):
        _i = i
        time.sleep(i)
        while True:
            task = self.q_in.get()
            if task is None:
                break
            n = self.process_path(_i, task)
            self.q_out.put(n)

    def load_database(self):
        start = datetime.datetime.now()

        n_proc = 5
        if sys.platform == "win32":
            n_proc = 1  # deactivate multiprocessing for window

        if n_proc == 1:
            count = 0
            for path in self.tasks:
                count += self.process_path(0, path)
        else:
            assert n_proc > 1, n_proc

            self.q_in = Queue()
            self.q_out = Queue()

            workers = []
            for i in range(n_proc):
                proc = Process(target=self.worker, args=(i,))
                proc.start()
                workers.append(proc)

            for path in self.tasks:
                self.q_in.put(path)

            for i in range(n_proc):
                self.q_in.put(None)

            count = 0
            for _ in progress_bar(
                desc="Indexing...",
                iterable=self.tasks,
                total=len(self.tasks),
            ):
                count += self.q_out.get()

            for p in workers:
                p.join()

            del self.q_in
            del self.q_out

        self._new_db().build_indexes()

        end = datetime.datetime.now()
        print(f"Indexed {plural(count,'field')} in {seconds(end - start)}.")

    def process_path(self, i, path):
        db = self._new_db()

        if db.already_loaded(self._format_path(path), self):
            LOG.warning(f"Skipping {path}, already loaded")
            return 0

        lst = []
        LOG.debug(f"Parsing file {path}")

        try:
            for field in _index_grib_file(
                path,
                with_statistics=self.with_statistics,
                position=i + 1,
            ):
                field["_path"] = self._format_path(path)
                lst.append(field)
        except PermissionError as e:
            LOG.error(f"Could not read {path}: {e}")
            return 0
        except Exception as e:
            LOG.exception(f"(grib-parsing) Ignoring {path}, {e}")
            return 0

        if not lst:
            LOG.warn(f"No entry found in {path}.")
            return 0

        return db.load_iterator(lst)

    @property
    def tasks(self):
        if self._tasks is not None:
            return self._tasks

        LOG.debug(f"Parsing files in {self.directory}")
        assert os.path.exists(self.directory), f"{self.directory} does not exist"
        assert os.path.isdir(self.directory), f"{self.directory} is not a directory"

        tasks = []
        for root, _, files in os.walk(self.directory, followlinks=self.followlinks):
            for name in files:
                path = os.path.join(root, name)
                _, ext = os.path.splitext(path)
                if ext not in self.extensions:
                    continue
                tasks.append(path)
        tasks = sorted(tasks)

        if tasks:
            if self.verbose:
                print(f"Found {len(tasks)} files to index.")
        else:
            LOG.error(f"Could not find any files to index in {self.directory}")

        self._tasks = tasks

        return self.tasks

    def _format_path(self, path):
        return {
            None: lambda x: x,
            True: lambda x: os.path.relpath(x, self.directory),
            False: lambda x: os.path.abspath(x),
        }[self.relative_paths](path)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging

LOG = logging.getLogger(__name__)


def normalize_selection(*args, **kwargs):
    from climetlab.arguments.transformers import ALL

    _kwargs = {}
    for a in args:
        if a is None:
            continue
        if isinstance(a, dict):
            _kwargs.update(a)
            continue
        raise ValueError(f"Cannot make a selection with {a}")

    _kwargs.update(kwargs)

    for k, v in _kwargs.items():
        assert (
            v is None
            or v is ALL
            or callable(v)
            or isinstance(v, (list, tuple, set))
            or isinstance(v, (str, int, float, datetime.datetime))
        ), f"Unsupported type: {type(v)} for key {k}"
    return _kwargs
